(***********************************************************************)
(*                                                                     *)
(*                                OCaml                                *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1999 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(***********************************************************************)

ouvre Asttypes
ouvre Location
ouvre Longident
ouvre Parsetree

module StringSet = Set.Make(struct type t = string soit compare = compare fin)

(* Collect free module identifiers in the a.s.t. *)

soit free_structure_names = ref StringSet.empty

soit rec addmodule bv lid =
  filtre lid avec
    Lident s ->
      si not (StringSet.mem s bv)
      alors free_structure_names := StringSet.add s !free_structure_names
  | Ldot(l, s) -> addmodule bv l
  | Lapply(l1, l2) -> addmodule bv l1; addmodule bv l2

soit add bv lid =
  filtre lid.txt avec
    Ldot(l, s) -> addmodule bv l
  | _ -> ()

soit addmodule bv lid = addmodule bv lid.txt

soit rec add_type bv ty =
  filtre ty.ptyp_desc avec
    Ptyp_any -> ()
  | Ptyp_var v -> ()
  | Ptyp_arrow(_, t1, t2) -> add_type bv t1; add_type bv t2
  | Ptyp_tuple tl -> List.iter (add_type bv) tl
  | Ptyp_constr(c, tl) -> add bv c; List.iter (add_type bv) tl
  | Ptyp_object (fl, _) -> List.iter (fonc (_, t) -> add_type bv t) fl
  | Ptyp_class(c, tl) -> add bv c; List.iter (add_type bv) tl
  | Ptyp_alias(t, s) -> add_type bv t
  | Ptyp_variant(fl, _, _) ->
      List.iter
        (fonction Rtag(_,_,stl) -> List.iter (add_type bv) stl
          | Rinherit sty -> add_type bv sty)
        fl
  | Ptyp_poly(_, t) -> add_type bv t
  | Ptyp_package pt -> add_package_type bv pt
  | Ptyp_extension _ -> ()

et add_package_type bv (lid, l) =
  add bv lid;
  List.iter (add_type bv) (List.map (fonc (_, e) -> e) l)

soit add_opt add_fn bv = fonction
    None -> ()
  | Some x -> add_fn bv x

soit add_constructor_decl bv pcd =
  List.iter (add_type bv) pcd.pcd_args; Misc.may (add_type bv) pcd.pcd_res

soit add_type_declaration bv td =
  List.iter
    (fonc (ty1, ty2, _) -> add_type bv ty1; add_type bv ty2)
    td.ptype_cstrs;
  add_opt add_type bv td.ptype_manifest;
  soit add_tkind = fonction
    Ptype_abstract -> ()
  | Ptype_variant cstrs ->
      List.iter (add_constructor_decl bv) cstrs
  | Ptype_record lbls ->
      List.iter (fonc pld -> add_type bv pld.pld_type) lbls dans
  add_tkind td.ptype_kind

soit rec add_class_type bv cty =
  filtre cty.pcty_desc avec
    Pcty_constr(l, tyl) ->
      add bv l; List.iter (add_type bv) tyl
  | Pcty_signature { pcsig_self = ty; pcsig_fields = fieldl } ->
      add_type bv ty;
      List.iter (add_class_type_field bv) fieldl
  | Pcty_arrow(_, ty1, cty2) ->
      add_type bv ty1; add_class_type bv cty2
  | Pcty_extension _ -> ()

et add_class_type_field bv pctf =
  filtre pctf.pctf_desc avec
    Pctf_inherit cty -> add_class_type bv cty
  | Pctf_val(_, _, _, ty) -> add_type bv ty
  | Pctf_method(_, _, _, ty) -> add_type bv ty
  | Pctf_constraint(ty1, ty2) -> add_type bv ty1; add_type bv ty2
  | Pctf_extension _ -> ()

soit add_class_description bv infos =
  add_class_type bv infos.pci_expr

soit add_class_type_declaration = add_class_description

soit pattern_bv = ref StringSet.empty

soit rec add_pattern bv pat =
  filtre pat.ppat_desc avec
    Ppat_any -> ()
  | Ppat_var _ -> ()
  | Ppat_alias(p, _) -> add_pattern bv p
  | Ppat_interval _
  | Ppat_constant _ -> ()
  | Ppat_tuple pl -> List.iter (add_pattern bv) pl
  | Ppat_construct(c, op) -> add bv c; add_opt add_pattern bv op
  | Ppat_record(pl, _) ->
      List.iter (fonc (lbl, p) -> add bv lbl; add_pattern bv p) pl
  | Ppat_array pl -> List.iter (add_pattern bv) pl
  | Ppat_or(p1, p2) -> add_pattern bv p1; add_pattern bv p2
  | Ppat_constraint(p, ty) -> add_pattern bv p; add_type bv ty
  | Ppat_variant(_, op) -> add_opt add_pattern bv op
  | Ppat_type li -> add bv li
  | Ppat_lazy p -> add_pattern bv p
  | Ppat_unpack id -> pattern_bv := StringSet.add id.txt !pattern_bv
  | Ppat_extension _ -> ()

soit add_pattern bv pat =
  pattern_bv := bv;
  add_pattern bv pat;
  !pattern_bv

soit rec add_expr bv exp =
  filtre exp.pexp_desc avec
    Pexp_ident l -> add bv l
  | Pexp_constant _ -> ()
  | Pexp_let(rf, pel, e) ->
      soit bv = add_bindings rf bv pel dans add_expr bv e
  | Pexp_fun (_, opte, p, e) ->
      add_opt add_expr bv opte; add_expr (add_pattern bv p) e
  | Pexp_function pel ->
      add_cases bv pel
  | Pexp_apply(e, el) ->
      add_expr bv e; List.iter (fonc (_,e) -> add_expr bv e) el
  | Pexp_match(e, pel) -> add_expr bv e; add_cases bv pel
  | Pexp_try(e, pel) -> add_expr bv e; add_cases bv pel
  | Pexp_tuple el -> List.iter (add_expr bv) el
  | Pexp_construct(c, opte) -> add bv c; add_opt add_expr bv opte
  | Pexp_variant(_, opte) -> add_opt add_expr bv opte
  | Pexp_record(lblel, opte) ->
      List.iter (fonc (lbl, e) -> add bv lbl; add_expr bv e) lblel;
      add_opt add_expr bv opte
  | Pexp_field(e, fld) -> add_expr bv e; add bv fld
  | Pexp_setfield(e1, fld, e2) -> add_expr bv e1; add bv fld; add_expr bv e2
  | Pexp_array el -> List.iter (add_expr bv) el
  | Pexp_ifthenelse(e1, e2, opte3) ->
      add_expr bv e1; add_expr bv e2; add_opt add_expr bv opte3
  | Pexp_sequence(e1, e2) -> add_expr bv e1; add_expr bv e2
  | Pexp_while(e1, e2) -> add_expr bv e1; add_expr bv e2
  | Pexp_for( _, e1, e2, _, e3) ->
      add_expr bv e1; add_expr bv e2; add_expr bv e3
  | Pexp_coerce(e1, oty2, ty3) ->
      add_expr bv e1;
      add_opt add_type bv oty2;
      add_type bv ty3
  | Pexp_constraint(e1, ty2) ->
      add_expr bv e1;
      add_type bv ty2
  | Pexp_send(e, m) -> add_expr bv e
  | Pexp_new li -> add bv li
  | Pexp_setinstvar(v, e) -> add_expr bv e
  | Pexp_override sel -> List.iter (fonc (s, e) -> add_expr bv e) sel
  | Pexp_letmodule(id, m, e) ->
      add_module bv m; add_expr (StringSet.add id.txt bv) e
  | Pexp_assert (e) -> add_expr bv e
  | Pexp_lazy (e) -> add_expr bv e
  | Pexp_poly (e, t) -> add_expr bv e; add_opt add_type bv t
  | Pexp_object { pcstr_self = pat; pcstr_fields = fieldl } ->
      soit bv = add_pattern bv pat dans List.iter (add_class_field bv) fieldl
  | Pexp_newtype (_, e) -> add_expr bv e
  | Pexp_pack m -> add_module bv m
  | Pexp_open (_ovf, m, e) -> addmodule bv m; add_expr bv e
  | Pexp_extension _ -> ()

et add_cases bv cases =
  List.iter (add_case bv) cases

et add_case bv {pc_lhs; pc_guard; pc_rhs} =
  soit bv = add_pattern bv pc_lhs dans
  add_opt add_expr bv pc_guard;
  add_expr bv pc_rhs

et add_bindings recf bv pel =
  soit bv' = List.fold_left (fonc bv x -> add_pattern bv x.pvb_pat) bv pel dans
  soit bv = si recf = Recursive alors bv' sinon bv dans
  List.iter (fonc x -> add_expr bv x.pvb_expr) pel;
  bv'

et add_modtype bv mty =
  filtre mty.pmty_desc avec
    Pmty_ident l -> add bv l
  | Pmty_alias l -> addmodule bv l
  | Pmty_signature s -> add_signature bv s
  | Pmty_functor(id, mty1, mty2) ->
      Misc.may (add_modtype bv) mty1;
      add_modtype (StringSet.add id.txt bv) mty2
  | Pmty_with(mty, cstrl) ->
      add_modtype bv mty;
      List.iter
        (fonction
          | Pwith_type (_, td) -> add_type_declaration bv td
          | Pwith_module (_, lid) -> addmodule bv lid
          | Pwith_typesubst td -> add_type_declaration bv td
          | Pwith_modsubst (_, lid) -> addmodule bv lid
        )
        cstrl
  | Pmty_typeof m -> add_module bv m
  | Pmty_extension _ -> ()

et add_signature bv = fonction
    [] -> ()
  | item :: rem -> add_signature (add_sig_item bv item) rem

et add_sig_item bv item =
  filtre item.psig_desc avec
    Psig_value vd ->
      add_type bv vd.pval_type; bv
  | Psig_type dcls ->
      List.iter (add_type_declaration bv) dcls; bv
  | Psig_exception pcd ->
      add_constructor_decl bv pcd; bv
  | Psig_module pmd ->
      add_modtype bv pmd.pmd_type; StringSet.add pmd.pmd_name.txt bv
  | Psig_recmodule decls ->
      soit bv' =
        List.fold_right StringSet.add
                        (List.map (fonc pmd -> pmd.pmd_name.txt) decls) bv
      dans
      List.iter (fonc pmd -> add_modtype bv' pmd.pmd_type) decls;
      bv'
  | Psig_modtype x ->
      début filtre x.pmtd_type avec
        None -> ()
      | Some mty -> add_modtype bv mty
      fin;
      bv
  | Psig_open (_ovf, lid, _) ->
      addmodule bv lid; bv
  | Psig_include (mty, _) ->
      add_modtype bv mty; bv
  | Psig_class cdl ->
      List.iter (add_class_description bv) cdl; bv
  | Psig_class_type cdtl ->
      List.iter (add_class_type_declaration bv) cdtl; bv
  | Psig_attribute _ | Psig_extension _ ->
      bv

et add_module bv modl =
  filtre modl.pmod_desc avec
    Pmod_ident l -> addmodule bv l
  | Pmod_structure s -> ignore (add_structure bv s)
  | Pmod_functor(id, mty, modl) ->
      Misc.may (add_modtype bv) mty;
      add_module (StringSet.add id.txt bv) modl
  | Pmod_apply(mod1, mod2) ->
      add_module bv mod1; add_module bv mod2
  | Pmod_constraint(modl, mty) ->
      add_module bv modl; add_modtype bv mty
  | Pmod_unpack(e) ->
      add_expr bv e
  | Pmod_extension _ ->
      ()

et add_structure bv item_list =
  List.fold_left add_struct_item bv item_list

et add_struct_item bv item =
  filtre item.pstr_desc avec
    Pstr_eval (e, _attrs) ->
      add_expr bv e; bv
  | Pstr_value(rf, pel) ->
      soit bv = add_bindings rf bv pel dans bv
  | Pstr_primitive vd ->
      add_type bv vd.pval_type; bv
  | Pstr_type dcls ->
      List.iter (add_type_declaration bv) dcls; bv
  | Pstr_exception pcd ->
      add_constructor_decl bv pcd; bv
  | Pstr_exn_rebind(id, l, _attrs) ->
      add bv l; bv
  | Pstr_module x ->
      add_module bv x.pmb_expr; StringSet.add x.pmb_name.txt bv
  | Pstr_recmodule bindings ->
      soit bv' =
        List.fold_right StringSet.add
          (List.map (fonc x -> x.pmb_name.txt) bindings) bv dans
      List.iter
        (fonc x -> add_module bv' x.pmb_expr)
        bindings;
      bv'
  | Pstr_modtype x ->
      début filtre x.pmtd_type avec
        None -> ()
      | Some mty -> add_modtype bv mty
      fin;
      bv
  | Pstr_open (_ovf, l, _attrs) ->
      addmodule bv l; bv
  | Pstr_class cdl ->
      List.iter (add_class_declaration bv) cdl; bv
  | Pstr_class_type cdtl ->
      List.iter (add_class_type_declaration bv) cdtl; bv
  | Pstr_include (modl, _attrs) ->
      add_module bv modl; bv
  | Pstr_attribute _ | Pstr_extension _ ->
      bv

et add_use_file bv top_phrs =
  ignore (List.fold_left add_top_phrase bv top_phrs)

et add_implementation bv l =
  ignore (add_structure bv l)

et add_top_phrase bv = fonction
  | Ptop_def str -> add_structure bv str
  | Ptop_dir (_, _) -> bv

et add_class_expr bv ce =
  filtre ce.pcl_desc avec
    Pcl_constr(l, tyl) ->
      add bv l; List.iter (add_type bv) tyl
  | Pcl_structure { pcstr_self = pat; pcstr_fields = fieldl } ->
      soit bv = add_pattern bv pat dans List.iter (add_class_field bv) fieldl
  | Pcl_fun(_, opte, pat, ce) ->
      add_opt add_expr bv opte;
      soit bv = add_pattern bv pat dans add_class_expr bv ce
  | Pcl_apply(ce, exprl) ->
      add_class_expr bv ce; List.iter (fonc (_,e) -> add_expr bv e) exprl
  | Pcl_let(rf, pel, ce) ->
      soit bv = add_bindings rf bv pel dans add_class_expr bv ce
  | Pcl_constraint(ce, ct) ->
      add_class_expr bv ce; add_class_type bv ct
  | Pcl_extension _ -> ()

et add_class_field bv pcf =
  filtre pcf.pcf_desc avec
    Pcf_inherit(_, ce, _) -> add_class_expr bv ce
  | Pcf_val(_, _, Cfk_concrete (_, e))
  | Pcf_method(_, _, Cfk_concrete (_, e)) -> add_expr bv e
  | Pcf_val(_, _, Cfk_virtual ty)
  | Pcf_method(_, _, Cfk_virtual ty) -> add_type bv ty
  | Pcf_constraint(ty1, ty2) -> add_type bv ty1; add_type bv ty2
  | Pcf_initializer e -> add_expr bv e
  | Pcf_extension _ -> ()

et add_class_declaration bv decl =
  add_class_expr bv decl.pci_expr
